/* This file was automatically generated by pyNLControl.*/


#include <stdlib.h>
#include <math.h>
#include "qpOASES.hpp"
#include "{{FileName}}_EVAL_CODE_Call.h"

void {{FileName}}_Call(double *xGuess, {%for xin in InName%}double *{{xin}}, {%endfor%} double *xOpt, double *Obj);