/* This file was automatically generated by pyNLControl.*/


#include "{{FileName}}.h"

void {{FileName}}_Call_Func({%for xin in Func_In%}casadi_real *{{xin}}, {%endfor%}{%for xout in Func_Out%}casadi_real *{{xout}}{% if not loop.last %}, {% endif %} {%endfor%});